//
//  SolidsViewController.m
//  Solids
//
//  Created by Michael Heyeck on 2/5/11.
//  Copyright 2011 Fair Oaks Labs, Inc. All rights reserved.
//

#import <QuartzCore/QuartzCore.h>

#import "SolidsViewController.h"
#import "EAGLView.h"
#import "glux.h"


// Uniform index.
enum {
    UNIFORM_TRANSLATE,
    NUM_UNIFORMS
};
GLint uniforms[NUM_UNIFORMS];

// Attribute index.
enum {
    ATTRIB_VERTEX,
    ATTRIB_COLOR,
    NUM_ATTRIBUTES
};


// Generic polyhedron parameters
#define MAX_FACE_POINTS	5
#define MAX_FACE_COUNT	20

// Data structure for face -> polyhedron construction
struct FacePos
{
	GLfloat angle;
	int		axis;
};
typedef struct FacePos FacePos;

// Generic polyhedron
struct Polyhedron
{
	// Face geometry
	int		frontSubfaceVertices;
	GLfloat vertices[MAX_FACE_POINTS*18];
	GLfloat normals[MAX_FACE_POINTS*18];
	
	// Face -> Solid conversion
	int		nPositions;
	FacePos positions[MAX_FACE_COUNT/2];
    GLfloat mirrorAxis[3];
};
typedef struct Polyhedron Polyhedron;


#define COPY_POINT(dst, dst_i, src, src_i) memcpy(dst+(dst_i)*3, src+(src_i)*3, sizeof(GLfloat)*3)


@interface SolidsViewController ()

@property (nonatomic, retain) EAGLContext *context;
@property (nonatomic, assign) CADisplayLink *displayLink;

- (BOOL)loadShaders;
- (BOOL)compileShader:(GLuint *)shader type:(GLenum)type file:(NSString *)file;
- (BOOL)linkProgram:(GLuint)prog;
- (BOOL)validateProgram:(GLuint)prog;

- (void)renderOctoInnerRadius:(float)ir outerRadius:(float)or explode:(float)x;
- (void)renderCubeInnerRadius:(float)ir outerRadius:(float)or explode:(float)x;
- (void)renderTetraInnerRadius:(float)ir outerRadius:(float)or explode:(float)x;
- (void)renderIcoInnerRadius:(float)ir outerRadius:(float)or explode:(float)x;
- (void)renderDodecaInnerRadius:(float)ir outerRadius:(float)or explode:(float)x;

- (BOOL)buildOctohedron:(Polyhedron*)polyhedron innerRadius:(float)ir outerRadius:(float)or;
- (BOOL)buildSexahedron:(Polyhedron*)polyhedron innerRadius:(float)ir outerRadius:(float)or;
- (BOOL)buildTetrahedron:(Polyhedron*)polyhedron innerRadius:(float)ir outerRadius:(float)or;
- (BOOL)buildIcosahedron:(Polyhedron*)polyhedron innerRadius:(float)ir outerRadius:(float)or;
- (BOOL)buildDodecahedron:(Polyhedron*)polyhedron innerRadius:(float)ir outerRadius:(float)or;

- (BOOL)buildPolyhedron:(Polyhedron*)polyhedron
			innerRadius:(float)ir
			outerRadius:(float)or
			  nVertices:(int)nVertices
		   baseVertices:(const GLfloat*)baseVertices
			baseNormals:(const GLfloat*)baseNormals
			 nPositions:(int)nPositions
		  facePositions:(const FacePos*)facePositions
			 mirrorAxis:(const GLfloat*)mirrorAxis;

- (void)renderPolyhedron:(Polyhedron*)polyhedron explodeRadius:(float)x;

@end


@implementation SolidsViewController

@synthesize animating, context, displayLink;

- (void)awakeFromNib
{
    EAGLContext *aContext = nil;//[[EAGLContext alloc] initWithAPI:kEAGLRenderingAPIOpenGLES2];
    
    if (!aContext)
    {
        aContext = [[EAGLContext alloc] initWithAPI:kEAGLRenderingAPIOpenGLES1];
    }
    
    if (!aContext)
        NSLog(@"Failed to create ES context");
    else if (![EAGLContext setCurrentContext:aContext])
        NSLog(@"Failed to set ES context current");
    
	self.context = aContext;
	[aContext release];
	
    [(EAGLView *)self.view setContext:context];
    [(EAGLView *)self.view setFramebuffer];
    
    if ([context API] == kEAGLRenderingAPIOpenGLES2)
        [self loadShaders];

	if (!polyhedra)
	{
		polyhedra = (Polyhedron*) malloc(sizeof(Polyhedron)*5);
		[self buildOctohedron:polyhedra+0 innerRadius:0.8f outerRadius:1.0f];
		[self buildSexahedron:polyhedra+1 innerRadius:0.8f outerRadius:1.0f];
		[self buildTetrahedron:polyhedra+2 innerRadius:0.8f outerRadius:1.0f];
		[self buildIcosahedron:polyhedra+3 innerRadius:0.8f outerRadius:1.0f];
		[self buildDodecahedron:polyhedra+4 innerRadius:0.8f outerRadius:1.0f];
		numPolyhedra = 5;
	}
	currentPolyhedron = 4;
	
    animating = FALSE;
    animationFrameInterval = 1;
    self.displayLink = nil;
}

- (void)dealloc
{
    if (program)
    {
        glDeleteProgram(program);
        program = 0;
    }
    
    // Tear down context.
    if ([EAGLContext currentContext] == context)
        [EAGLContext setCurrentContext:nil];
    
    [context release];
    
	if (polyhedra)
	{
		free(polyhedra);
		polyhedra = nil;
	}
	
    [super dealloc];
}

- (void)viewWillAppear:(BOOL)animated
{
    [self startAnimation];
    
    [super viewWillAppear:animated];
}

- (void)viewWillDisappear:(BOOL)animated
{
    [self stopAnimation];
    
    [super viewWillDisappear:animated];
}

- (void)viewDidUnload
{
	[super viewDidUnload];
	
    if (program)
    {
        glDeleteProgram(program);
        program = 0;
    }

    // Tear down context.
    if ([EAGLContext currentContext] == context)
        [EAGLContext setCurrentContext:nil];
	self.context = nil;	
}

- (NSInteger)animationFrameInterval
{
    return animationFrameInterval;
}

- (void)setAnimationFrameInterval:(NSInteger)frameInterval
{
    /*
	 Frame interval defines how many display frames must pass between each time the display link fires.
	 The display link will only fire 30 times a second when the frame internal is two on a display that refreshes 60 times a second. The default frame interval setting of one will fire 60 times a second when the display refreshes at 60 times a second. A frame interval setting of less than one results in undefined behavior.
	 */
    if (frameInterval >= 1)
    {
        animationFrameInterval = frameInterval;
        
        if (animating)
        {
            [self stopAnimation];
            [self startAnimation];
        }
    }
}

- (void)startAnimation
{
    if (!animating)
    {
        CADisplayLink *aDisplayLink = [CADisplayLink displayLinkWithTarget:self selector:@selector(drawFrame)];
        [aDisplayLink setFrameInterval:animationFrameInterval];
        [aDisplayLink addToRunLoop:[NSRunLoop currentRunLoop] forMode:NSDefaultRunLoopMode];
        self.displayLink = aDisplayLink;
        
        animating = TRUE;
    }
}

- (void)stopAnimation
{
    if (animating)
    {
        [self.displayLink invalidate];
        self.displayLink = nil;
        animating = FALSE;
    }
}

- (void)drawFrame
{
    [(EAGLView *)self.view setFramebuffer];
    
    // Replace the implementation of this method to do your own custom drawing.
    static const GLfloat squareVertices[] = {
        -0.5f, -0.33f,
         0.5f, -0.33f,
        -0.5f,  0.33f,
         0.5f,  0.33f,
    };
    
    static const GLubyte squareColors[] = {
        255, 255,   0, 255,
        0,   255, 255, 255,
        0,     0,   0,   0,
        255,   0, 255, 255,
    };
    
    static const GLfloat lightPos[] = {
		-6.0f, 6.0f, 8.0f, 1.0f,
		 6.0f, 0.0f, -8.0f, 1.0f
	};
    static const GLfloat lightPow[] = {
		0.7f, 0.3f, 0.7f, 1.0f,
		0.3f, 0.25f, 0.3f, 1.0f
	};
    static const GLfloat cameraPos[] = {
		6.0f, 6.0f, 8.0f
	};
	
	static float transY = 0.0f;
    
    glClearColor(0.45f, 0.45f, 0.45f, 1.0f);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    
    if ([context API] == kEAGLRenderingAPIOpenGLES2)
    {
        // Use shader program.
        glUseProgram(program);
        
        // Update uniform value.
        glUniform1f(uniforms[UNIFORM_TRANSLATE], (GLfloat)transY);
        transY += 0.075f;	
        
        // Update attribute values.
        glVertexAttribPointer(ATTRIB_VERTEX, 2, GL_FLOAT, 0, 0, squareVertices);
        glEnableVertexAttribArray(ATTRIB_VERTEX);
        glVertexAttribPointer(ATTRIB_COLOR, 4, GL_UNSIGNED_BYTE, 1, 0, squareColors);
        glEnableVertexAttribArray(ATTRIB_COLOR);
        
        // Validate program before drawing. This is a good check, but only really necessary in a debug build.
        // DEBUG macro must be defined in your debug configurations if that's not already the case.
#if defined(DEBUG)
        if (![self validateProgram:program])
        {
            NSLog(@"Failed to validate program: %d", program);
            return;
        }
#endif
		// Render original demo
		glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
    }
    else
    {
        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();
		gluPerspective(20.0f, [(EAGLView*)self.view aspect], 1.5f, 50.0f);
		glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();
		glLightfv(GL_LIGHT0, GL_POSITION, lightPos);
		glLightfv(GL_LIGHT1, GL_POSITION, lightPos+4);
		gluLookAt(cameraPos[0], cameraPos[1], cameraPos[2],
				  0.0f, 0.0f, 0.0f,
				  0.0f, 1.0f, 0.0);
        
		glEnable(GL_BLEND);
		glEnable(GL_CULL_FACE);
		glEnable(GL_DEPTH_TEST);
		glEnable(GL_LIGHTING);
		glEnable(GL_LIGHT0);
		glEnable(GL_LIGHT1);
		glEnable(GL_COLOR_MATERIAL);
		
		glLightfv(GL_LIGHT0, GL_DIFFUSE, lightPow);
		glLightfv(GL_LIGHT1, GL_DIFFUSE, lightPow+4);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		
		UIColor* color = [UIColor colorWithHue:sinf(transY*.1)/2.0f+0.5f saturation:1.0f brightness:1.0f alpha:1.0f];
		const CGFloat* components = CGColorGetComponents([color CGColor]);
		glColor4f(components[0], components[1], components[2], components[3]);
		
		glRotatef(transY*180.0/M_PI/4, 1.0f, 1.0f, -1.0f);
		[self renderPolyhedron:polyhedra+currentPolyhedron explodeRadius:sinf(transY)/2.0f+0.5f];

        transY += 0.075f;
    }
	
    [(EAGLView *)self.view presentFramebuffer];
}

- (void)didReceiveMemoryWarning
{
    // Releases the view if it doesn't have a superview.
    [super didReceiveMemoryWarning];
    
    // Release any cached data, images, etc. that aren't in use.
}

- (BOOL)compileShader:(GLuint *)shader type:(GLenum)type file:(NSString *)file
{
    GLint status;
    const GLchar *source;
    
    source = (GLchar *)[[NSString stringWithContentsOfFile:file encoding:NSUTF8StringEncoding error:nil] UTF8String];
    if (!source)
    {
        NSLog(@"Failed to load vertex shader");
        return FALSE;
    }
    
    *shader = glCreateShader(type);
    glShaderSource(*shader, 1, &source, NULL);
    glCompileShader(*shader);
    
#if defined(DEBUG)
    GLint logLength;
    glGetShaderiv(*shader, GL_INFO_LOG_LENGTH, &logLength);
    if (logLength > 0)
    {
        GLchar *log = (GLchar *)malloc(logLength);
        glGetShaderInfoLog(*shader, logLength, &logLength, log);
        NSLog(@"Shader compile log:\n%s", log);
        free(log);
    }
#endif
    
    glGetShaderiv(*shader, GL_COMPILE_STATUS, &status);
    if (status == 0)
    {
        glDeleteShader(*shader);
        return FALSE;
    }
    
    return TRUE;
}

- (BOOL)linkProgram:(GLuint)prog
{
    GLint status;
    
    glLinkProgram(prog);
    
#if defined(DEBUG)
    GLint logLength;
    glGetProgramiv(prog, GL_INFO_LOG_LENGTH, &logLength);
    if (logLength > 0)
    {
        GLchar *log = (GLchar *)malloc(logLength);
        glGetProgramInfoLog(prog, logLength, &logLength, log);
        NSLog(@"Program link log:\n%s", log);
        free(log);
    }
#endif
    
    glGetProgramiv(prog, GL_LINK_STATUS, &status);
    if (status == 0)
        return FALSE;
    
    return TRUE;
}

- (BOOL)validateProgram:(GLuint)prog
{
    GLint logLength, status;
    
    glValidateProgram(prog);
    glGetProgramiv(prog, GL_INFO_LOG_LENGTH, &logLength);
    if (logLength > 0)
    {
        GLchar *log = (GLchar *)malloc(logLength);
        glGetProgramInfoLog(prog, logLength, &logLength, log);
        NSLog(@"Program validate log:\n%s", log);
        free(log);
    }
    
    glGetProgramiv(prog, GL_VALIDATE_STATUS, &status);
    if (status == 0)
        return FALSE;
    
    return TRUE;
}

- (BOOL)loadShaders
{
    GLuint vertShader, fragShader;
    NSString *vertShaderPathname, *fragShaderPathname;
    
    // Create shader program.
    program = glCreateProgram();
    
    // Create and compile vertex shader.
    vertShaderPathname = [[NSBundle mainBundle] pathForResource:@"Shader" ofType:@"vsh"];
    if (![self compileShader:&vertShader type:GL_VERTEX_SHADER file:vertShaderPathname])
    {
        NSLog(@"Failed to compile vertex shader");
        return FALSE;
    }
    
    // Create and compile fragment shader.
    fragShaderPathname = [[NSBundle mainBundle] pathForResource:@"Shader" ofType:@"fsh"];
    if (![self compileShader:&fragShader type:GL_FRAGMENT_SHADER file:fragShaderPathname])
    {
        NSLog(@"Failed to compile fragment shader");
        return FALSE;
    }
    
    // Attach vertex shader to program.
    glAttachShader(program, vertShader);
    
    // Attach fragment shader to program.
    glAttachShader(program, fragShader);
    
    // Bind attribute locations.
    // This needs to be done prior to linking.
    glBindAttribLocation(program, ATTRIB_VERTEX, "position");
    glBindAttribLocation(program, ATTRIB_COLOR, "color");
    
    // Link program.
    if (![self linkProgram:program])
    {
        NSLog(@"Failed to link program: %d", program);
        
        if (vertShader)
        {
            glDeleteShader(vertShader);
            vertShader = 0;
        }
        if (fragShader)
        {
            glDeleteShader(fragShader);
            fragShader = 0;
        }
        if (program)
        {
            glDeleteProgram(program);
            program = 0;
        }
        
        return FALSE;
    }
    
    // Get uniform locations.
    uniforms[UNIFORM_TRANSLATE] = glGetUniformLocation(program, "translate");
    
    // Release vertex and fragment shaders.
    if (vertShader)
        glDeleteShader(vertShader);
    if (fragShader)
        glDeleteShader(fragShader);
    
    return TRUE;
}


- (void)renderOctoInnerRadius:(float)ir outerRadius:(float)or explode:(float)x
{
	Polyhedron p;
	if ([self buildOctohedron:&p innerRadius:ir outerRadius:or])
	{
		[self renderPolyhedron:&p explodeRadius:x];
	}
}


- (void)renderCubeInnerRadius:(float)ir outerRadius:(float)or explode:(float)x
{
	Polyhedron p;
	if ([self buildSexahedron:&p innerRadius:ir outerRadius:or])
	{
		[self renderPolyhedron:&p explodeRadius:x];
	}
}


- (void)renderTetraInnerRadius:(float)ir outerRadius:(float)or explode:(float)x
{
	Polyhedron p;
	if ([self buildTetrahedron:&p innerRadius:ir outerRadius:or])
	{
		[self renderPolyhedron:&p explodeRadius:x];
	}
}


- (void)renderIcoInnerRadius:(float)ir outerRadius:(float)or explode:(float)x
{
	Polyhedron p;
	if ([self buildIcosahedron:&p innerRadius:ir outerRadius:or])
	{
		[self renderPolyhedron:&p explodeRadius:x];
	}
}


- (void)renderDodecaInnerRadius:(float)ir outerRadius:(float)or explode:(float)x
{
	Polyhedron p;
	if ([self buildDodecahedron:&p innerRadius:ir outerRadius:or])
	{
		[self renderPolyhedron:&p explodeRadius:x];
	}
}


- (BOOL)buildOctohedron:(Polyhedron*)polyhedron innerRadius:(float)ir outerRadius:(float)or
{
	// Front subface vertices
    static const GLfloat baseVertices[] = {
		 0.00000000000000000, -1.00000000000000000,  0.00000000000000000,
		 0.00000000000000000,  0.00000000000000000,  1.00000000000000000,
		-1.00000000000000000,  0.00000000000000000,  0.00000000000000000
	};
	
	// All face normals
    static const GLfloat baseNormals[] = {
		-0.57735026918962584, -0.57735026918962584,  0.57735026918962584,
		 0.57735026918962584,  0.57735026918962584, -0.57735026918962584,
		 1.00000000000000000,  0.00000000000000000,  0.00000000000000000,
		 0.00000000000000000,  1.00000000000000000,  0.00000000000000000,
		 0.00000000000000000,  0.00000000000000000, -1.00000000000000000
	};
	
	// Face -> Solid conversion
	static const FacePos facePositions[] = {
		  0.0f, 2,
		 90.0f, 2,
		180.0f, 2,
		270.0f, 2
	};
	
	// Mirroring
    static const GLfloat mirrorAxis[] = {0.0f, 1.0f, 0.0f};
	
	// Subcontracting
	return [self buildPolyhedron:polyhedron
					 innerRadius:ir
					 outerRadius:or
					   nVertices:3
					baseVertices:baseVertices
					 baseNormals:baseNormals
					  nPositions:4
				   facePositions:facePositions
					  mirrorAxis:mirrorAxis];
}


- (BOOL)buildSexahedron:(Polyhedron*)polyhedron innerRadius:(float)ir outerRadius:(float)or
{
	// Front subface vertices
    static const GLfloat baseVertices[] = {
		-0.57735026918962584, -0.57735026918962584,  0.57735026918962584,
		 0.57735026918962584, -0.57735026918962584,  0.57735026918962584,
		 0.57735026918962584,  0.57735026918962584,  0.57735026918962584,
		-0.57735026918962584,  0.57735026918962584,  0.57735026918962584
	};
	
	// All face normals
    static const GLfloat baseNormals[] = {
		 0.00000000000000000,  0.00000000000000000,  1.00000000000000000,
		-0.00000000000000000, -0.00000000000000000, -1.00000000000000000,
		-0.00000000000000000, -0.70710678118654746, -0.70710678118654746,
		 0.70710678118654746,  0.00000000000000000, -0.70710678118654746,
		 0.00000000000000000,  0.70710678118654746, -0.70710678118654746,
		-0.70710678118654746,  0.00000000000000000, -0.70710678118654746	
	};
	
	// Face -> Solid conversion
	static const FacePos facePositions[] = {
		  0.0f, 2,
		120.0f, 2,
		240.0f, 2
	};
	
	// Mirroring
    static const GLfloat mirrorAxis[] = {-0.57735026918962584,  0.0f, 0.57735026918962584};
	
	// Subcontracting
	return [self buildPolyhedron:polyhedron
					 innerRadius:ir
					 outerRadius:or
					   nVertices:4
					baseVertices:baseVertices
					 baseNormals:baseNormals
					  nPositions:3
				   facePositions:facePositions
					  mirrorAxis:mirrorAxis];
}


- (BOOL)buildTetrahedron:(Polyhedron*)polyhedron innerRadius:(float)ir outerRadius:(float)or
{
	// Front subface vertices
    static const GLfloat baseVertices[] = {
		 0.00000000000000000, -0.33333333333333331, -0.94280904158206347,
		 0.81649658092772603, -0.33333333333333331,  0.47140452079103184,
		-0.81649658092772603, -0.33333333333333331,  0.47140452079103184
	};
	
	// All face normals
    static const GLfloat baseNormals[] = {
		 0.00000000000000000, -1.00000000000000000,  0.00000000000000000,
		-0.00000000000000000,  1.00000000000000000, -0.00000000000000000,
		 0.50000000000000011,  0.81649658092772592, -0.28867513459481287,
		-0.00000000000000000,  0.81649658092772615,  0.57735026918962562,
		-0.50000000000000011,  0.81649658092772592, -0.28867513459481287
	};
	
	// Face -> Solid conversion
	static const FacePos facePositions[] = {
		  0.0f,	0,
		120.0f,	0
	};
	
	// Mirroring
    static const GLfloat mirrorAxis[] = {0.0f, -0.33333333333333331, 0.47140452079103184};
	
	// Subcontracting
	return [self buildPolyhedron:polyhedron
					 innerRadius:ir
					 outerRadius:or
					   nVertices:3
					baseVertices:baseVertices
					 baseNormals:baseNormals
					  nPositions:2
				   facePositions:facePositions
					  mirrorAxis:mirrorAxis];
}


- (BOOL)buildIcosahedron:(Polyhedron*)polyhedron innerRadius:(float)ir outerRadius:(float)or
{
	// Front subface vertices
    static const GLfloat baseVertices[] = {
		0.00000000000000000,  1.00000000000000000,  0.00000000000000000,
		0.00000000000000000,  0.44721359549995798,  0.89442719099991586,
		0.85065080835203988,  0.44721359549995798,  0.27639320225002106
	};
	
	// All face normals
    static const GLfloat baseNormals[] = {
		 0.35682208977308993,  0.79465447229176622,  0.49112347318842309,
		-0.35682208977308993, -0.79465447229176622, -0.49112347318842309,
		-1.00000000000000000,  0.00000000000000000, -0.00000000000000000,
		 0.30901699437494745, -0.85065080835203988,  0.42532540417602005,
		 0.30901699437494751,  0.00000000000000000, -0.95105651629515364
	};
	
	// Face -> Solid conversion
	static const FacePos facePositions[] = {
		  0.0f,	0,
		 72.0f,	0,
		144.0f,	0,
		216.0f,	0,
		288.0f,	0,
		144.0f,	1,
		216.0f,	1,
		288.0f,	1,
		144.0f,	2,
		216.0f,	2
	};
	
	// Mirroring
    static const GLfloat mirrorAxis[] = {0.68819096023558668, 0.0f, -0.5f};
	
	// Subcontracting
	return [self buildPolyhedron:polyhedron
					 innerRadius:ir
					 outerRadius:or
					   nVertices:3
					baseVertices:baseVertices
					 baseNormals:baseNormals
					  nPositions:10
				   facePositions:facePositions
					  mirrorAxis:mirrorAxis];
}


- (BOOL)buildDodecahedron:(Polyhedron*)polyhedron innerRadius:(float)ir outerRadius:(float)or
{
	// Front subface vertices
    static const GLfloat baseVertices[] = {
		 0.00000000000000000,  0.79465447229176600,  0.60706199820668627,
		 0.57735026918962573,  0.79465447229176600,  0.18759247408507992,
		 0.35682208977308999,  0.79465447229176600, -0.49112347318842309,
		-0.35682208977308999,  0.79465447229176600, -0.49112347318842309,
		-0.57735026918962573,  0.79465447229176600,  0.18759247408507992
	};
	
	// All face normals
    static const GLfloat baseNormals[] = {
		 0.00000000000000000,  1.00000000000000000,  0.00000000000000000,
		-0.00000000000000000, -1.00000000000000000, -0.00000000000000000,
		 0.50000000000000000, -0.52573111211913370,  0.68819096023558679,
		 0.80901699437494734, -0.52573111211913359, -0.26286555605956663,
		 0.00000000000000000, -0.52573111211913370, -0.85065080835203977,
		-0.80901699437494734, -0.52573111211913359, -0.26286555605956663,
		-0.50000000000000000, -0.52573111211913370,  0.68819096023558679		
	};
	
	// Face -> Solid conversion
	static const FacePos facePositions[] = {
		  0.0f, 0,
		120.0f,	0,
		240.0f, 0,
		120.0f, 2,
		240.0f, 2,
		120.0f, 4
	};
	
	// Mirroring
    static const GLfloat mirrorAxis[] = {1.0f, 0.0f, 0.0f};
	
	// Subcontracting
	return [self buildPolyhedron:polyhedron
					 innerRadius:ir
					 outerRadius:or
					   nVertices:5
					baseVertices:baseVertices
					 baseNormals:baseNormals
					  nPositions:6
				   facePositions:facePositions
					  mirrorAxis:mirrorAxis];
}


- (BOOL)buildPolyhedron:(Polyhedron*)polyhedron
			innerRadius:(float)ir
			outerRadius:(float)or
			  nVertices:(int)nVertices
		   baseVertices:(const GLfloat*)baseVertices
			baseNormals:(const GLfloat*)baseNormals
			 nPositions:(int)nPositions
		  facePositions:(const FacePos*)facePositions
			 mirrorAxis:(const GLfloat*)mirrorAxis
{
	// Bounds check
	if ((!polyhedron)					||
		(ir <= 0.0f)					||
		(or <= ir)						||
		(nVertices < 3)					||
		(nVertices > MAX_FACE_POINTS)	||
		(!baseVertices)					||
		(!baseNormals)					||
		(nPositions < 1)				||
		(nPositions > MAX_FACE_COUNT/2)	||
		(!facePositions)				||
		(!mirrorAxis)					)
	{
		return FALSE;
	}
	
	// Helpers
	CGFloat* frontVertices	= polyhedron->vertices;
	CGFloat* backVertices	= frontVertices + (2*nVertices-1)*3;
	CGFloat* edgeVertices	= backVertices + 3;
	CGFloat* frontNormals	= polyhedron->normals;
	CGFloat* backNormals	= frontNormals + (2*nVertices-1)*3;
	CGFloat* edgeNormals	= backNormals + 3;
	
	// Build a full set of face points from the given front subface
	// (Scaling from the unit sphere to the given radii)
	polyhedron->frontSubfaceVertices = nVertices;
	for (int v = 0; v < nVertices; v++)
	{
		int i0 = v*3, i1 = -i0;
		
		frontVertices[i0+0] = baseVertices[i0+0] * or;
		frontVertices[i0+1] = baseVertices[i0+1] * or;
		frontVertices[i0+2] = baseVertices[i0+2] * or;
		
		backVertices[i1+0] = baseVertices[i0+0] * ir;
		backVertices[i1+1] = baseVertices[i0+1] * ir;
		backVertices[i1+2] = baseVertices[i0+2] * ir;
		
		COPY_POINT(frontNormals, i0/3, baseNormals, 0);
		COPY_POINT(backNormals, i1/3, baseNormals, 1);
	}
	for (int e = 0; e < nVertices; e++)
	{
		int v0 = (e+1)%nVertices, v1 = e;
		
		COPY_POINT(edgeVertices, e*4+0, frontVertices, v0);
		COPY_POINT(edgeVertices, e*4+1, frontVertices, v1);
		COPY_POINT(edgeVertices, e*4+2, backVertices, -v0);
		COPY_POINT(edgeVertices, e*4+3, backVertices, -v1);
		
		COPY_POINT(edgeNormals, e*4+0, baseNormals, e+2);
		COPY_POINT(edgeNormals, e*4+1, baseNormals, e+2);
		COPY_POINT(edgeNormals, e*4+2, baseNormals, e+2);
		COPY_POINT(edgeNormals, e*4+3, baseNormals, e+2);
	}
	
	// Setup face -> solid conversion
	polyhedron->nPositions = nPositions;
	memcpy(polyhedron->positions, facePositions, nPositions*sizeof(FacePos));
	memcpy(polyhedron->mirrorAxis, mirrorAxis, sizeof(polyhedron->mirrorAxis));
	
	// Success
	return TRUE;
}


- (void)renderPolyhedron:(Polyhedron*)polyhedron explodeRadius:(float)x
{
	GLfloat* axis;
	
	// This is 1.1-specific rendering code
    if ([context API] != kEAGLRenderingAPIOpenGLES1) return;
	
	// Bounds check
	if (x < 0.0f) return;
	
	// Prepare to render
	glVertexPointer(3, GL_FLOAT, 0, polyhedron->vertices);
	glEnableClientState(GL_VERTEX_ARRAY);
	glNormalPointer(GL_FLOAT, 0, polyhedron->normals);
	glEnableClientState(GL_NORMAL_ARRAY);
	
	// Render faces
	int fv = polyhedron->frontSubfaceVertices;
    for (int i = 0; i < 2; i++)
	{
		if (i)
		{
			axis = polyhedron->mirrorAxis;
			
			glPushMatrix();
			glRotatef(180.0f, axis[0], axis[1], axis[2]);
		}
		
		for (int j = 0; j < polyhedron->nPositions; j++)
		{
			axis = polyhedron->vertices + polyhedron->positions[j].axis*3;
			
			glPushMatrix();
			glRotatef(polyhedron->positions[j].angle, axis[0], axis[1], axis[2]);
			
			if (x > 0.0f)
			{
				// When exploded, draw all subfaces
				axis = polyhedron->normals;
				glTranslatef(x*axis[0], x*axis[1], x*axis[2]);
				
				glDrawArrays(GL_TRIANGLE_FAN, 0, fv);
				glDrawArrays(GL_TRIANGLE_FAN, fv, fv);
				
				for (int k = 0; k < fv; k++)
					glDrawArrays(GL_TRIANGLE_STRIP, 2*fv+4*k, 4);
			}
			else
			{
				// When solid, draw the front subface only
				glDrawArrays(GL_TRIANGLE_FAN, 0, fv);
			}
			
			glPopMatrix();
		}
		
		if (i) glPopMatrix();
	}
}


- (void)tapDetectingView:(TapDetectingView*)view gotSingleTapAtPoint:(CGPoint)tapPoint
{
	currentPolyhedron = (currentPolyhedron+1)%numPolyhedra;
}


- (void)tapDetectingView:(TapDetectingView*)view gotDoubleTapAtPoint:(CGPoint)tapPoint
{
	currentPolyhedron = (currentPolyhedron+1)%numPolyhedra;
}

@end
